pub enum HeaderOption {
Normal(u8),
Compressed(u8),
}Expand description
HeaderOption to pick when encoding FIT file, this will optimize the size of the resulting FIT file.
Available options: Normal(0-15), Compressed(0-3).
Default: Normal(0)
Variants
Normal(u8)
Set Normal Header with the number of maximum message definition interleave allowed. Valid value: 0-15;
Compressed(u8)
Set Compressed Header with the number of maximum message definition interleave allowed. Valid value: 0-3;