Decoder

Struct Decoder 

pub struct Decoder { /* private fields */ }
Decoder for decoding FIT file.

Implementations

Source

impl Decoder

pub const fn new() -> Self

Create new Decoder for decoding FIT file. For more options, use Decoder::builder() to build the Decoder.

pub const fn builder() -> Builder

Create new Decoder with options for decoding FIT file.

pub fn stream<'a, R>(&'a mut self, reader: R) -> Stream<'a, R>

Creates a Stream from a mutably borrowed Decoder for streaming decoding of the given reader.

Example:

// ...
let mut dec = Decoder::new();
let mut stream = dec.stream(&mut reader);

while let Some(event) = stream.next() {
   // do something with event having borrowed data
}
pub fn decode<R>(&mut self, reader: R) -> Result<Option<FIT>, Error<R::Error>>
where R: Read,

Decode returns a single FIT sequence. If it’s a chained FIT file, call this method multiple times until it return Ok(None) or Err(err).

pub fn release_heap_memory(&mut self)

Releases heap memory used by the Decoder back to the allocator, so it can be reused by subsequent allocations in memory-constrained environments such as embedded systems when a statically allocated Decoder is no longer in use.