pub struct Encoder { /* private fields */ }Expand description
Encoder for encoding FIT file.
Implementations
Sourceimpl Encoder
impl Encoder
Sourcepub const fn new() -> Encoder
pub const fn new() -> Encoder
Create new Encoder for encoding FIT file.
For more options, use Encoder::builder() to build the Encoder.
Sourcepub fn stream<'a, W>(&'a mut self, writer: W) -> Stream<'a, W>
pub fn stream<'a, W>(&'a mut self, writer: W) -> Stream<'a, W>
Creates a Stream from a mutably borrowed Encoder for streaming encoding of the given writer.
Sourcepub fn encode<W>(
&mut self,
writer: W,
fit: &mut FIT,
) -> Result<(), Error<W::Error>>
pub fn encode<W>( &mut self, writer: W, fit: &mut FIT, ) -> Result<(), Error<W::Error>>
Encode the given fit to the writer.
Sourcepub fn release_heap_memory(&mut self)
pub fn release_heap_memory(&mut self)
Releases heap memory used by the Encoder back to the allocator, so it can be reused
by subsequent allocations in memory-constrained environments such as embedded systems
when a statically allocated Encoder is no longer in use.