StreamEncoder

Struct StreamEncoder 

pub struct StreamEncoder<'a, W> { /* private fields */ }
A Stream from a mutably borrowed Encoder for streaming encoding.

Implementations

Source

impl<'a, W: Write + Seek> Stream<'a, W>

pub fn write_message( &mut self, mesg: &mut Message, ) -> Result<(), Error<W::Error>>

Write message to the writer. When done writing all messages, call finish() to complete this FIT sequence.

pub fn finish(&mut self) -> Result<(), Error<W::Error>>

Finish the current FIT sequence by finalizing the correct FileHeader.

Calling write_message() after this point will write message for the next FIT sequence on the same writer, and finish() MUST be called again.